from shirley.interfaces import chat
from shirley.interfaces.chat import Chat
from shirley.interfaces.footer import Footer
from shirley.interfaces.header import Header
from shirley.interfaces.tts import TextToSpeech
