from shirley.options.client import ClientOptions, ChatClientOptions, TextToSpeechClientOptions
from shirley.options.interfaces import (
    InterfaceOptions,
    HeaderInterfaceOptions,
    FooterInterfaceOptions,
    ChatInterfaceOptions,
    TextToSpeechInterfaceOptions
)
from shirley.options.components import ChatbotOptions
